#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import fcntl
import getopt
import os
import re
import signal
import subprocess
import sys
import time
from socket import gethostname

from urllib.parse import urlencode
from urllib.request import urlopen

# The pCache Version
pcacheversion = "4.2.3"

# Log message levels
DEBUG, INFO, WARN, ERROR = "DEBUG", "INFO ", "WARN ", "ERROR"

# filename for locking
LOCK_NAME = ".LOCK"

# Session ID
sessid = "%s.%s" % (int(time.time()), os.getpid())


# Run a command with a timeout
def run_cmd(args, timeout=0):

    class Alarm(Exception):
        pass

    def alarm_handler(signum, frame):
        raise Alarm

    # Execute the command as a subprocess
    try:
        p = subprocess.Popen(args=args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)

    except:
        return (-2, None)

    # Set the timer if a timeout value was given
    if (timeout > 0):
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(timeout)

    # Wait for the command to complete
    try:

        # Collect the output when the command completes
        stdout = p.communicate()[0][:-1]

        # Commmand completed in time, cancel the alarm
        if (timeout > 0):
            signal.alarm(0)

    # Command timed out
    except Alarm:

        # The pid of our spawn
        pids = [p.pid]

        # The pids of the spawn of our spawn
        pids.extend(get_process_children(p.pid))

        # Terminate all of the evil spawn
        for pid in pids:
            try:
                os.kill(pid, signal.SIGKILL)
            except OSError:
                pass

        # Return a timeout error
        return (-1, None)

    return (p.returncode, stdout)


def get_process_children(pid):

    # Get a list of all pids assocaited with a given pid
    p = subprocess.Popen(args='ps --no-headers -o pid --ppid %d' % pid,
                         shell=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    # Wait and fetch the stdout
    stdout, stderr = p.communicate()

    # Return a list of pids as tuples
    return [int(pr) for pr in stdout.split()]


def unitize(x):

    suff = 'BKMGTPEZY'

    while ((x >= 1024) and suff):
        x /= 1024.0
        suff = suff[1:]
    return "%.4g%s" % (x, suff[0])


class Pcache:

    def Usage(self):
        msg = """Usage: %s [flags] copy_prog [copy_flags] input output""" % self.progname
        sys.stderr.write("%s\n" % msg)  # py3, py2
    #  print>>sys.stderr, "  flags are: "
    #  "s:r:m:Cy:A:R:t:r:g:fFl:VvqpH:S:",
    #  "scratch-dir=",
    #  "storage-root=",
    #  "max-space=",
    #  "clean",
    #  "hysterisis=",
    #  "accept=",
    #  "reject=",
    #  "timeout=",
    #  "retry=",
    #  "force",
    #  "flush-cache",
    #  "guid=",
    #  "log=",
    #  "version",
    #  "verbose",
    #  "debug",
    #  "quiet",
    #  "panda",
    #  "hostname",
    #  "sitename"

    def __init__(self):
        os.umask(0)
        self.storage_root = "/pnfs"
        self.scratch_dir = "/scratch/"
        self.pcache_dir = self.scratch_dir + "pcache/"
        self.log_file = self.pcache_dir + "pcache.log"
        self.max_space = "80%"
        self.percent_max = None
        self.bytes_max = None
        # self.max_space = "10T"
        self.hysterisis = 0.75
        # self.hysterisis = 0.9
        self.clean = False
        self.transfer_timeout = "600"
        self.max_retries = 3
        self.guid = None
        self.accept_patterns = []
        self.reject_patterns = []
        self.force = False
        self.flush = False
        self.verbose = False
        self.quiet = False
        self.debug = False
        self.hostname = None
        self.sitename = None  # XXX can we get this from somewhere?
        self.update_panda = False
        self.panda_url = "https://pandaserver.cern.ch:25443/server/panda/"
        self.local_src = None
        self.skip_download = False

        # internal variables
        self.sleep_interval = 15
        self.force = False
        self.locks = {}
        self.deleted_guids = []
        self.version = pcacheversion

    def parse_args(self, args):
        # handle pcache flags and leave the rest in self.args

        try:
            opts, args = getopt.getopt(args,
                                       "s:x:m:Cy:A:R:t:r:g:fFl:VvdqpPH:S:L:X:",
                                       ["scratch-dir=",
                                        "storage-root=",
                                        "max-space=",
                                        "clean",
                                        "hysterisis=",
                                        "accept=",
                                        "reject=",
                                        "timeout=",
                                        "retry=",
                                        "force",
                                        "flush-cache",
                                        "guid=",
                                        "log=",
                                        "version",
                                        "verbose",
                                        "debug",
                                        "quiet",
                                        "print-stats",
                                        "panda",
                                        "hostname",
                                        "sitename",
                                        "local-src"])

            # XXXX cache, stats, reset, clean, delete, inventory
            # TODO: move checksum/size validation from lsm to pcache
        except getopt.GetoptError as err:
            sys.stderr.write("%s\n" % str(err))
            self.Usage()
            self.fail(100)

        for opt, arg in opts:
            if opt in ("-s", "--scratch-dir"):
                self.scratch_dir = arg
                # Make sure scratch_dir endswith /
                if not self.scratch_dir.endswith("/"):
                    self.scratch_dir += "/"
                self.pcache_dir = self.scratch_dir + "pcache/"
                self.log_file = self.pcache_dir + "pcache.log"
            elif opt in ("-x", "--storage-root"):
                self.storage_root = arg
            elif opt in ("-m", "--max-space"):
                self.max_space = arg
            elif opt in ("-y", "--hysterisis"):
                if arg.endswith('%'):
                    self.hysterisis = float(arg[:-1]) / 100
                else:
                    self.hysterisis = float(arg)
            elif opt in ("-A", "--accept"):
                self.accept_patterns.append(arg)
            elif opt in ("-R", "--reject"):
                self.reject_patterns.append(arg)
            elif opt in ("-t", "--timeout"):
                self.transfer_timeout = arg
            elif opt in ("-f", "--force"):
                self.force = True
            elif opt in ("-F", "--flush-cache"):
                self.flush = True
            elif opt in ("-C", "--clean"):
                self.clean = True
            elif opt in ("-g", "--guid"):
                if arg == 'None':
                    self.guid = None
                else:
                    self.guid = arg
            elif opt in ("-r", "--retry"):
                self.max_retries = int(arg)
            elif opt in ("-V", "--version"):
                print((str(self.version)))
                sys.exit(0)
            elif opt in ("-l", "--log"):
                self.log_file = arg
            elif opt in ("-v", "--verbose"):
                self.verbose = True
            elif opt in ("-d", "--debug"):
                self.debug = True
            elif opt in ("-q", "--quiet"):
                self.quiet = True
            elif opt in ("-p", "--print-stats"):
                self.print_stats()
                sys.exit(0)
            elif opt in ("-P", "--panda"):
                self.update_panda = True
            elif opt in ("-H", "--hostname"):
                self.hostname = arg
            elif opt in ("-S", "--sitename"):
                self.sitename = arg
            elif opt in ("-L", "--local-src"):
                self.local_src = str(arg)
            elif opt in ("-X", "--skip-download"):
                if str(arg) in ('True', 'true') or arg:
                    self.skip_download = True

        # Treatment of limits on pcache size
        self._convert_max_space()

        # Convert timeout to seconds
        t = self.transfer_timeout
        mult = 1
        suff = t[-1]
        if suff in ('H', 'h'):
            mult = 3600
            t = t[:-1]
        elif suff in ('M', 'm'):
            mult = 60
            t = t[:1]
        elif suff in ('S', 's'):
            mult = 1
            t = t[:-1]
        self.transfer_timeout = mult * int(t)

        # Pre-compile regexes
        self.accept_patterns = list(map(re.compile, self.accept_patterns))
        self.reject_patterns = list(map(re.compile, self.reject_patterns))

        # Set host and name
        if self.hostname is None:
            self.hostname = gethostname()
        if self.sitename is None:
            self.sitename = os.environ.get("SITE", "")  # XXXX

        # All done
        self.args = args

    def _convert_max_space(self):
        '''
        Added by Rucio team. Converts max allowed space usage of pcache into units used by this tool.
        :input self.max_space: limit set by user
        :output self.percent_max: max percentage of pcache space used
        :output self.bytes_max: max size in bytes of pcache space used
        '''

        # Convert max_space arg to percent_max or bytes_max
        if self.max_space.endswith('%'):
            self.percent_max = float(self.max_space[:-1])
            self.bytes_max = None
        else:  # handle suffix
            self.percent_max = None
            m = self.max_space.upper()
            index = "BKMGTPEZY".find(m[-1])
            if index >= 0:
                self.bytes_max = float(m[:-1]) * (1024**index)
            else:  # Numeric value w/o units (exception if invalid)
                self.bytes_max = float(self.max_space)

    def clean_pcache(self, max_space=None):
        '''
        Added by Rucio team. Cleans pcache in case it is over limit.
        Used for tests of the pcache functionality. Can be called without other init.
        '''

        self.t0 = time.time()
        self.progname = "pcache"

        # set max. occupancy of pcache:
        if max_space:
            self.max_space = max_space
        self._convert_max_space()

        # Fail on extra args
        if not self.scratch_dir:
            self.Usage()
            self.fail(100)

        # hardcoded pcache dir
        self.pcache_dir = self.scratch_dir + '/pcache/'

        # clean pcache
        self.maybe_start_cleaner_thread()

    def check_and_link(self, src='', dst='', dst_prefix='', scratch_dir='/scratch/', storage_root=None, force=False,
                       guid=None, log_file=None, version='', hostname=None, sitename=None, local_src=None):
        '''
        Added by Rucio team. Replacement for the main method.
        Checks whether a file is in pcache:
         - if yes: creates a hardlink to the file in pcahce
         - if not:
              - returns 500 and leaves it to Rucio
              - Rucio downloads a file
        Makes hardlink in pcache to downloaded file:
         - needs :param local_src: path to downloaded file
        '''
        self.t0 = time.time()
        self.progname = "pcache"
        self.pcache_dir = scratch_dir + '/pcache/'
        self.src = src
        self.dst = dst
        self.dst_prefix = dst_prefix
        self.sitename = sitename
        self.hostname = hostname
        self.guid = guid
        if log_file:
            self.log_file = log_file
        self.local_src = local_src
        self.version = version
        self.storage_root = storage_root

        # Cache dir may have been wiped
        if ((not os.path.exists(self.pcache_dir)) and self.update_panda):
            self.panda_flush_cache()

        # Create the pCache directory
        if (self.mkdir_p(self.pcache_dir)):
            self.fail(101)

        self.log(INFO, "%s %s invoked as: API", self.progname, self.version)

        # Fail on extra args
        if not scratch_dir:
            self.Usage()
            self.fail(100)

        # If the source is lfn:, execute original command, no further action
        if (self.src.startswith('lfn:')):
            # status = os.execvp(self.copy_util, self.args)
            os._exit(1)

        # If the destination is a local file, do some rewrites
        if (self.dst.startswith('file:')):
            self.dst_prefix = 'file:'
            self.dst = self.dst[5:]
            # Leave one '/' on dst
            while ((len(self.dst) > 1) and (self.dst[1] == '/')):
                self.dst_prefix += '/'
                self.dst = self.dst[1:]

        # load file into pcache
        self.create_pcache_dst_dir()
        # XXXX TODO _ dst_dir can get deleted before lock!
        waited = False

        # If another transfer is active, lock_dir will block
        if (self.lock_dir(self.pcache_dst_dir, blocking=False)):
            waited = True
            self.log(INFO, "%s locked, waiting", self.pcache_dst_dir)
            self.lock_dir(self.pcache_dst_dir, blocking=True)

        if (waited):
            self.log(INFO, "waited %.2f secs", time.time() - self.t0)

        if force:
            self.empty_dir(self.pcache_dst_dir)

        # The name of the cached version of this file
        cache_file = self.pcache_dst_dir + "data"

        # Check if the file is in cache or we need to transfer it down
        if (os.path.exists(cache_file)):
            exit_status = 0
            copy_status = None
            self.log(INFO, "check_and_link: file found in cache")
            self.log(INFO, "cache hit %s", self.src)
            self.update_stats("cache_hits")
            self.finish()
            if (os.path.exists(self.dst)):
                copy_status = 1
        elif self.local_src:
            exit_status = 1
            copy_status = None
            self.log(INFO, "check_and_link: local replica found, linking to pcache")
            self.finish()
        else:
            self.log(INFO, "check_and_link: %s file not found in pcache and was not downloaded yet", self.src)
            return (500, None)

        self.unlock_dir(self.pcache_dst_dir)
        self.log(INFO, "total time %.2f secs", time.time() - self.t0)

        # in case that the pcache is over limit
        self.maybe_start_cleaner_thread()

        # Return if the file was cached, copied or an error (and its code)
        return (exit_status, copy_status)

    def main(self, args):

        # args
        self.cmdline = ' '.join(args)
        self.t0 = time.time()
        self.progname = args[0] or "pcache"

        # Must have a list of arguments
        if (self.parse_args(args[1:])):
            self.Usage()
            self.fail(100)

        # Cache dir may have been wiped
        if ((not os.path.exists(self.pcache_dir)) and self.update_panda):
            self.panda_flush_cache()

        # Create the pCache directory
        if (self.mkdir_p(self.pcache_dir)):
            self.fail(101)

        self.log(INFO, "%s %s invoked as: %s", self.progname, self.version, self.cmdline)

        # Are we flushing the cache
        if (self.flush):
            if (self.args):
                sys.stderr.write("--flush not compatible with other options")
                self.fail(100)
            else:
                self.flush_cache()
                sys.exit(0)

        # Are we cleaning the cache
        if (self.clean):
            # size = self.do_cache_inventory()
            self.maybe_start_cleaner_thread()
            if (len(self.args) < 1):
                sys.exit(0)

        # Fail on extra args
        if (len(self.args) < 3):
            self.Usage()
            self.fail(100)

        self.copy_util = self.args[0]
        self.copy_args = self.args[1:-2]
        self.src = self.args[-2]
        self.dst = self.args[-1]
        self.dst_prefix = ''

        # If the source is lfn:, execute original command, no further action
        if (self.src.startswith('lfn:')):
            # status = os.execvp(self.copy_util, self.args)
            os._exit(1)

        # If the destination is a local file, do some rewrites
        if (self.dst.startswith('file:')):
            self.dst_prefix = 'file:'
            self.dst = self.dst[5:]
            # Leave one '/' on dst
            while ((len(self.dst) > 1) and (self.dst[1] == '/')):
                self.dst_prefix += '/'
                self.dst = self.dst[1:]

        # Execute original command, no further action
        if (not (self.dst.startswith(self.scratch_dir) and self.accept(self.src) and (not self.reject(self.src)))):
            os.execvp(self.copy_util, self.args)
            os._exit(1)

        # XXXX todo:  fast-path - try to acquire lock
        # first, if that succeeds, don't call
        # create_pcache_dst_dir

        # load file into pcache
        self.create_pcache_dst_dir()
        # XXXX TODO _ dst_dir can get deleted before lock!
        waited = False

        # If another transfer is active, lock_dir will block
        if (self.lock_dir(self.pcache_dst_dir, blocking=False)):
            waited = True
            self.log(INFO, "%s locked, waiting", self.pcache_dst_dir)
            self.lock_dir(self.pcache_dst_dir, blocking=True)

        if (waited):
            self.log(INFO, "waited %.2f secs", time.time() - self.t0)

        if (self.force):
            self.empty_dir(self.pcache_dst_dir)

        # The name of the cached version of this file
        cache_file = self.pcache_dst_dir + "data"

        # Check if the file is in cache or we need to transfer it down
        if (os.path.exists(cache_file)):
            exit_status = 1
            copy_status = None
            self.log(INFO, "cache hit %s", self.src)
            self.update_stats("cache_hits")
            self.finish()
        else:
            if self.skip_download:
                return (500, None)
            self.update_stats("cache_misses")
            exit_status, copy_status = self.pcache_copy_in()
            if ((exit_status == 0) or (exit_status == 2)):
                self.finish()

        self.unlock_dir(self.pcache_dst_dir)
        self.log(INFO, "total time %.2f secs", time.time() - self.t0)

        self.maybe_start_cleaner_thread()

        # Return if the file was cached, copied or an error (and its code)
        return (exit_status, copy_status)

    def finish(self, local_src=None):
        cache_file = self.pcache_dst_dir + "data"
        self.update_mru()
        if self.local_src:
            if (self.make_hard_link(self.local_src, cache_file)):
                self.fail(102)
        else:
            if (self.make_hard_link(cache_file, self.dst)):
                self.fail(102)

    def pcache_copy_in(self):

        cache_file = self.pcache_dst_dir + "data"

        # Record source URL
        try:
            fname = self.pcache_dst_dir + "src"
            f = open(fname, 'w')
            f.write(self.src + '\n')
            f.close()
            self.chmod(fname, 0o666)
        except:
            pass

        # Record GUID if given
        if (self.guid):
            try:
                fname = self.pcache_dst_dir + "guid"
                f = open(fname, 'w')
                f.write(self.guid + '\n')
                f.close()
                self.chmod(fname, 0o666)
            except:
                pass

        # Try to transfer the file up the the number of retries allowed
        retry = 0
        while (retry <= self.max_retries):

            # Is this is a retry attempt, log it as such
            if (retry > 0):
                self.log(INFO, "do_transfer: retry %s", retry)

            # Do the transfer. exit_status will be either
            # 0 - success
            # 3 - Transfer command failed. copy_status has the return code
            # 4 - Transfer command timed out
            # 5 - Transfer command was not found
            exit_status, copy_status = self.do_transfer()

            # If success, stop trying, otherwise increment the retry count
            if (exit_status == 0):
                break
            retry += 1

        # Did the transfer succeed
        if (exit_status == 0):

            # If we succeeded on a retry, return status 2 and the retries
            if (retry == 0):
                copy_status = None
            else:
                exit_status = 2
                copy_status = retry

            # Update the cache information
            if self.local_src:
                self.update_cache_size(os.stat(self.local_src).st_size)
            else:
                self.update_cache_size(os.stat(cache_file).st_size)

            # Update the panda cache
            if (self.guid and self.update_panda):
                self.panda_add_cache_files((self.guid,))

        return (exit_status, copy_status)

    def create_pcache_dst_dir(self):

        d = self.src
        index = d.find(self.storage_root)

        if (index >= 0):
            d = d[index:]
        else:
            index = d.find("SFN=")
            if (index >= 0):
                d = d[index + 4:]

        # self.log(INFO, '%s', self.storage_root)
        # self.log(INFO, '%s', d)
        # XXXX any more patterns to look for?
        d = os.path.normpath(self.pcache_dir + "CACHE/" + d)
        if (not d.endswith('/')):
            d += '/'

        self.pcache_dst_dir = d
        status = self.mkdir_p(d)
        if (status):
            self.log(ERROR, "mkdir %s %s", d, status)
            self.fail(103)

    def get_disk_usage(self):
        p = os.popen("df -P %s | tail -1" % self.pcache_dir, 'r')
        data = p.read()
        status = p.close()
        if status:
            self.log(ERROR, "get_disk_usage: df command failed, status=%s", status)
            sys.exit(1)
        tok = data.split()
        percent = tok[-2]
        if not percent.endswith('%'):
            self.log(ERROR, "get_disk_usage: cannot parse df output: %s", data)
            sys.exit(1)
        percent = int(percent[:-1])
        return percent

    def over_limit(self, factor=1.0):
        if self.percent_max:
            return self.get_disk_usage() > factor * self.percent_max
        if self.bytes_max:
            return self.get_cache_size() > factor * self.bytes_max
        return False

    def clean_cache(self):
        t0 = time.time()

        self.log(INFO, "starting cleanup, cache size=%s, usage=%s%%",
                 unitize(self.get_cache_size()),
                 self.get_disk_usage())

        for link in self.list_by_mru():
            try:
                d = os.readlink(link)

            except OSError as e:
                self.log(ERROR, "readlink: %s", e)
                continue

            self.log(DEBUG, "deleting %s", d)

            if os.path.exists(d):
                self.empty_dir(d)
            else:
                self.log(WARN, "Attempt to delete missing file %s", d)
                self.flush_cache()
                break

            # empty_dir should also delete MRU symlink, but
            # mop up here in there is some problem with the
            # backlink

            try:
                os.unlink(link)

            except OSError as e:
                if e.errno != errno.ENOENT:
                    self.log(ERROR, "unlink: %s", e)

            if not self.over_limit(self.hysterisis):
                break

        self.log(INFO, "cleanup complete, cache size=%s, usage=%s%%, time=%.2f secs",
                 self.get_cache_size(),
                 self.get_disk_usage(),
                 time.time() - t0)

    def list_by_mru(self):
        mru_dir = self.pcache_dir + "MRU/"
        for root, dirs, files in os.walk(mru_dir):
            dirs.sort()
            for d in dirs:
                path = os.path.join(root, d)
                if os.path.islink(path):
                    dirs.remove(d)
                    yield path
            if files:
                files.sort()
                for file in files:
                    path = os.path.join(root, file)
                    yield path

    def flush_cache(self):
        # Delete everything in CACHE, MRU, and reset stats
        self.log(INFO, "flushing cache")
        if self.update_panda:
            self.panda_flush_cache()
        self.reset_stats()
        ts = '.' + str(time.time())
        for d in "CACHE", "MRU":
            d = self.pcache_dir + d
            try:
                os.rename(d, d + ts)
                os.system("rm -rf %s &" % (d + ts))
            except OSError as e:
                if e.errno != errno.ENOENT:
                    self.log(ERROR, "%s: %s", d, e)

    def do_transfer(self):

        # Cache file and transfer file locations
        cache_file = self.pcache_dst_dir + "data"
        xfer_file = self.pcache_dst_dir + "xfer"

        # Remove any transfer file with the same name
        try:
            os.unlink(xfer_file)
        except OSError as e:
            if e.errno != errno.ENOENT:
                self.log(ERROR, "unlink: %s", e)

        # Build the copy command with the destination into the xfer location
        args = self.args[:]
        args[-1] = self.dst_prefix + xfer_file

        # Save the current time for timing output
        t0 = time.time()

        # Do the copy with a timeout
        if self.local_src:
            return(0, None)
        else:
            copy_status, copy_output = run_cmd(args, self.transfer_timeout)

        # Did the command timeout
        if (copy_status == -1):
            self.log(ERROR, "copy command timed out, elapsed time = %.2f sec", time.time() - t0)
            self.cleanup_failed_transfer()
            return (4, None)
        elif (copy_status == -2):
            self.log(ERROR, "copy command was not found")
            self.cleanup_failed_transfer()
            return (5, None)

        # Display any output from the copy
        if (copy_output):
            print('%s' % copy_output)

        # Did the copy succeed (good status and an existing file)
        if ((copy_status > 0) or (not os.path.exists(xfer_file))):
            self.log(ERROR, "copy command failed, elapsed time = %.2f sec", time.time() - t0)
            self.cleanup_failed_transfer()
            return (3, copy_status)

        self.log(INFO, "copy command succeeded, elapsed time = %.2f sec", time.time() - t0)

        try:
            os.rename(xfer_file, cache_file)
            # self.log(INFO, "rename %s %s", xfer_file, cache_file)
        except OSError:  # Fatal error if we can't do this
            self.log(ERROR, "rename %s %s", xfer_file, cache_file)
            try:
                os.unlink(xfer_file)
            except:
                pass
            self.fail(104)

        # Make the file readable to all
        self.chmod(cache_file, 0o666)

        # Transfer completed, return the transfer command status
        return (0, None)

    def maybe_start_cleaner_thread(self):
        if not self.over_limit():
            return
        # exit immediately if another cleaner is active
        cleaner_lock = os.path.join(self.pcache_dir, ".clean")
        if self.lock_file(cleaner_lock, blocking=False):
            self.log(INFO, "cleanup not starting:  %s locked", cleaner_lock)
            return
        # see http://www.faqs.org/faqs/unix-faq/faq/part3/section-13.html
        # for explanation of double-fork
        pid = os.fork()
        if pid:  # parent
            os.waitpid(pid, 0)
            return
        else:  # child
            self.daemonize()
            pid = os.fork()
            if pid:
                os._exit(0)
            # grandchild
            self.clean_cache()
            self.unlock_file(cleaner_lock)
            os._exit(0)

    def make_hard_link(self, src, dst):
        self.log(INFO, "linking %s to %s", src, dst)
        try:
            if os.path.exists(dst):
                os.unlink(dst)
            os.link(src, dst)
        except OSError as e:
            self.log(ERROR, "make_hard_link: %s", e)
            ret = e.errno
            if ret == errno.ENOENT:
                try:
                    stat_info = os.stat(src)
                    self.log(INFO, "stat(%s) = %s", src, stat_info)
                except:
                    self.log(INFO, "cannot stat %s", src)
                try:
                    stat_info = os.stat(dst)
                    self.log(INFO, "stat(%s) = %s", dst, stat_info)
                except:
                    self.log(INFO, "cannot stat %s", dst)
            return ret

    def reject(self, name):
        for pat in self.reject_patterns:
            if pat.search(name):
                return True
        return False

    def accept(self, name):
        if not self.accept_patterns:
            return True
        for pat in self.accept_patterns:
            if pat.search(name):
                return True
        return False

    def get_stat(self, stats_dir, stat_name):
        filename = os.path.join(self.pcache_dir, stats_dir, stat_name)
        try:
            f = open(filename, 'r')
            data = int(f.read().strip())
            f.close()
        except:
            data = 0
        return data

    def print_stats(self):
        print(("Cache size: %s", unitize(self.get_stat("CACHE", "size"))))
        print(("Cache hits: %s", self.get_stat("stats", "cache_hits")))
        print(("Cache misses: %s", self.get_stat("stats", "cache_misses")))

    def reset_stats(self):
        stats_dir = os.path.join(self.pcache_dir, "stats")
        try:
            for f in os.listdir(stats_dir):
                try:
                    os.unlink(os.path.join(stats_dir, f))
                except:
                    pass
        except:
            pass
        # XXXX error handling
        pass

    def update_stat_file(self, stats_dir, name, delta):  # internal
        stats_dir = os.path.join(self.pcache_dir, stats_dir)
        self.mkdir_p(stats_dir)
        self.lock_dir(stats_dir)
        stats_file = os.path.join(stats_dir, name)
        try:
            f = open(stats_file, 'r')
            data = f.read()
            f.close()
            value = int(data)
        except:
            # XXXX
            value = 0
        value += delta
        try:
            f = open(stats_file, 'w')
            f.write("%s\n" % value)
            f.close()
            self.chmod(stats_file, 0o666)
        except:
            pass
            # XXX
        self.unlock_dir(stats_dir)

    def update_stats(self, name, delta=1):
        return self.update_stat_file("stats", name, delta)

    def update_cache_size(self, bytes_):
        return self.update_stat_file("CACHE", "size", bytes_)

    def get_cache_size(self):
        filename = os.path.join(self.pcache_dir, "CACHE", "size")
        size = 0

        try:
            f = open(filename)
            data = f.read()
            size = int(data)
        except:
            pass

        # If we could not fetch the size, do a reinventory
        if size == 0:
            size = self.do_cache_inventory()

        # The size should never be negative, so lets cleanup and start over
        if size < 0:
            self.log(WARN, "CACHE corruption found. Negative CACHE size: %d", size)
            self.flush_cache()
            size = 0

        return size

    def do_cache_inventory(self):

        inventory_lock = os.path.join(self.pcache_dir, ".inventory")
        if self.lock_file(inventory_lock, blocking=False):
            return

        size = 0

        self.log(INFO, "starting inventory")

        for root, dirs, files in os.walk(self.pcache_dir):
            for f in files:
                if f == "data":
                    fullname = os.path.join(root, f)
                    try:
                        size += os.stat(fullname).st_size
                    except OSError as e:
                        self.log(ERROR, "stat(%s): %s", fullname, e)

        filename = os.path.join(self.pcache_dir, "CACHE", "size")

        try:
            f = open(filename, 'w')
            f.write("%s\n" % size)
            f.close()
            self.chmod(filename, 0o666)
        except:
            pass  # XXXX

        self.unlock_file(inventory_lock)
        self.log(INFO, "inventory complete, cache size %s", size)
        return size

    def daemonize(self):
        if self.debug:
            return
        try:
            os.setsid()
        except OSError:
            pass
        try:
            os.chdir("/")
        except OSError:
            pass
        try:
            os.umask(0)
        except OSError:
            pass
        n = os.open("/dev/null", os.O_RDWR)
        i, o, e = sys.stdin.fileno(), sys.stdout.fileno(), sys.stderr.fileno()
        os.dup2(n, i)
        os.dup2(n, o)
        os.dup2(n, e)
        MAXFD = 1024
        try:
            import resource  # Resource usage information.
            maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
            if (maxfd == resource.RLIM_INFINITY):
                maxfd = MAXFD
        except:
            try:
                maxfd = os.sysconf("SC_OPEN_MAX")
            except:
                maxfd = MAXFD  # use default value

        for fd in range(0, maxfd + 1):
            try:
                os.close(fd)
            except:
                pass

    # Panda server callback functions
    def do_http_post(self, url, data):
        # see http://www.faqs.org/faqs/unix-faq/faq/part3/section-13.html
        # for explanation of double-fork (is it overkill here?)
        pid = os.fork()
        if pid:  # parent
            os.waitpid(pid, 0)
            return
        else:  # child
            self.daemonize()
            pid = os.fork()
            if pid:
                os._exit(0)
            # grandchild
            retry = 0
            # This will retry for up to 1 hour, at 2 minute intervals
            while retry < 30:
                try:
                    u = urlopen(url, data=urlencode(data))
                    ret = u.read()
                    u.close()
                    self.log(INFO, "http post to %s, retry %s, data='%s', return='%s'",
                             url, retry, data, ret)
                    if ret == "True":
                        break
                except:
                    exc, msg, tb = sys.exc_info()
                    self.log(ERROR, "post to %s, data=%s, error=%s", url, data, msg)
                retry += 1
                time.sleep(120)
            # finished, don't keep the child thread around!
            os._exit(0)

    def panda_flush_cache(self):
        self.do_http_post(self.panda_url + "flushCacheDB",
                          data={"site": self.sitename,
                                "node": self.hostname})

    def panda_add_cache_files(self, guids):
        self.do_http_post(self.panda_url + "addFilesToCacheDB",
                          data={"site": self.sitename,
                                "node": self.hostname,
                                "guids": ','.join(guids)})

    def panda_del_cache_files(self, guids):
        self.do_http_post(self.panda_url + "deleteFilesFromCacheDB",
                          data={"site": self.sitename,
                                "node": self.hostname,
                                "guids": ','.join(guids)})

    # Locking functions
    def lock_dir(self, d, create=True, blocking=True):
        lock_name = os.path.join(d, LOCK_NAME)
        lock_status = self.lock_file(lock_name, blocking)
        if (not lock_status):  # succeeded
            return
        if ((lock_status == errno.ENOENT) and create):
            mkdir_status = self.mkdir_p(d)
            if (mkdir_status):
                self.log(ERROR, "mkdir %s %s", d, mkdir_status)
                self.fail(105)
            lock_status = self.lock_file(lock_name, blocking)
        return lock_status

    def unlock_dir(self, d):
        return self.unlock_file(os.path.join(d, LOCK_NAME))

    def lock_file(self, name, blocking=True):
        if name in self.locks:
            self.log(DEBUG, "lock_file: %s already locked", name)
            return
        try:
            f = open(name, 'w')
        except IOError as e:
            self.log(ERROR, "open: %s", e)
            return e.errno

        self.locks[name] = f
        flag = fcntl.LOCK_EX
        if not blocking:
            flag |= fcntl.LOCK_NB
        while True:
            try:
                status = fcntl.lockf(f, flag)
                break
            except IOError as e:
                if e.errno in (errno.EAGAIN, errno.EACCES) and not blocking:
                    f.close()
                    del self.locks[name]
                    return e.errno
                if e.errno != errno.EINTR:
                    status = e.errno
                    self.log(ERROR, "lockf: %s", e)
                    self.fail(106)
        return status

    def unlock_file(self, name):
        f = self.locks.get(name)
        if not f:
            self.log(DEBUG, "unlock_file: %s not locked", name)
            return

        # XXXX does this create a possible race condition?
        if 0:
            try:
                os.unlink(name)
            except:
                pass
        status = fcntl.lockf(f, fcntl.LOCK_UN)
        f.close()
        del self.locks[name]
        return status

    def unlock_all(self):
        for filename, f in list(self.locks.items()):
            try:
                f.close()
                os.unlink(filename)
            except:
                pass

    # Cleanup functions
    def delete_file_and_parents(self, name):
        try:
            os.unlink(name)
        except OSError as e:
            if e.errno != errno.ENOENT:
                self.log(ERROR, "unlink: %s", e)
                self.fail(107)
        self.delete_parents_recursive(name)

    def delete_parents_recursive(self, name):  # internal
        try:
            dirname = os.path.dirname(name)
            if not os.listdir(dirname):
                os.rmdir(dirname)
                self.delete_parents_recursive(dirname)
        except OSError as e:
            self.log(DEBUG, "delete_parents_recursive: %s", e)

    def update_mru(self):
        now = time.time()
        link_to_mru = self.pcache_dst_dir + "mru"
        if os.path.exists(link_to_mru):
            link = os.readlink(link_to_mru)
            self.delete_file_and_parents(link)

        try:
            os.unlink(link_to_mru)
        except OSError as e:
            if e.errno != errno.ENOENT:
                self.log(ERROR, "unlink: %s", e)
                self.fail(108)

        mru_dir = self.pcache_dir + "MRU/" + time.strftime("%Y/%m/%d/%H/%M/",
                                                           time.localtime(now))

        self.mkdir_p(mru_dir)

        # getting symlink
        name = "%.3f" % (now % 60)
        ext = 0
        while True:
            if ext:
                link_from_mru = "%s%s-%s" % (mru_dir, name, ext)
            else:
                link_from_mru = "%s%s" % (mru_dir, name)

            try:
                os.symlink(self.pcache_dst_dir, link_from_mru)
                break
            except OSError as e:
                if e.errno == errno.EEXIST:
                    ext += 1  # add an extension & retry if file exists
                    continue
                else:
                    self.log(ERROR, "symlink: %s %s", e, link_from_mru)
                    self.fail(109)

        while True:
            try:
                os.symlink(link_from_mru, link_to_mru)
                break
            except OSError as e:
                if e.errno == errno.EEXIST:
                    try:
                        os.unlink(link_to_mru)
                    except OSError as e:
                        if e.errno != errno.ENOENT:
                            self.log(ERROR, "unlink: %s %s", e, link_to_mru)
                            self.fail(109)
                else:
                    self.log(ERROR, "symlink: %s %s", e, link_from_mru)
                    self.fail(109)

    def cleanup_failed_transfer(self):
        try:
            os.unlink(self.pcache_dir + 'xfer')
        except:
            pass

    def empty_dir(self, d):
        status = None
        bytes_deleted = 0
        for name in os.listdir(d):
            size = 0
            fullname = os.path.join(d, name)
            if name == "data":
                try:
                    size = os.stat(fullname).st_size
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        self.log(WARN, "stat: %s", e)
            elif name == "guid":
                try:
                    guid = open(fullname).read().strip()
                    self.deleted_guids.append(guid)
                except:
                    pass  # XXXX
            elif name == "mru" and os.path.islink(fullname):
                try:
                    mru_file = os.readlink(fullname)
                    os.unlink(fullname)
                    self.delete_file_and_parents(mru_file)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        self.log(WARN, "empty_dir: %s", e)
            try:
                if self.debug:
                    print(("UNLINK %s", fullname))
                os.unlink(fullname)
                bytes_deleted += size
            except OSError as e:
                if e.errno != errno.ENOENT:
                    self.log(WARN, "empty_dir2: %s", e)
                # self.fail()
        self.update_cache_size(-bytes_deleted)
        self.delete_parents_recursive(d)
        return status

    def chmod(self, path, mode):
        try:
            os.chmod(path, mode)
        except OSError as e:
            if e.errno != errno.EPERM:  # Cannot chmod files we don't own!
                self.log(ERROR, "chmod %s %s", path, e)

    def mkdir_p(self, d, mode=0o777):
        # Thread-safe
        try:
            os.makedirs(d, mode)
            return 0
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                # Don't use log here, log dir may not exist
                sys.stderr.write("%s\n" % str(e))
                return e.errno

    def log(self, level, msg, *args):

        # Disable all logging
        if (self.quiet):
            return

        if ((level == DEBUG) and (not self.debug)):
            return

        msg = "%s %s %s %s %s\n" % (time.strftime("%F %H:%M:%S"),
                                    sessid,
                                    self.hostname,
                                    level,
                                    str(msg) % args)

        try:
            f = open(self.log_file, "a+", 0o666)
            f.write(msg)
            f.close()

        except Exception as e:
            sys.stderr.write("%s\n" % str(e))
            sys.stderr.write(msg)
            sys.stderr.flush()

        if (self.debug or self.verbose or (level == ERROR)):
            sys.stderr.write(msg)
            sys.stderr.flush()

    def fail(self, errcode=1):
        self.unlock_all()
        sys.exit(errcode)

##################################################################################

# pCache exit_status will be
#
#    0 - File was transferred into cache and is ready
#    1 - File is cached and ready to use
#    2 - File was transferred but with a retry (copy_status has the retry count)
#    3 - Transfer command failed (copy_status has the transfer return code)
#    4 - Transfer command timed out
#
#  100 - Usage error
#  101 - Cache directory does not exist
#  102 - Cache hard link error
#  103 - Cache destination mkdir error
#  104 - Cache rename error
#  105 - Cache locking error
#  106 - Cache file locking error
#  107 - Cache cleanup error
#  108 - Cache MRU update error
#  109 - Cache MRU link error
#  500 - Is file in pcache? No other action


if (__name__ == "__main__"):

    # Load pCache
    p = Pcache()

    # Save the passed arguments
    args = sys.argv

    # Find the file
    exit_status, copy_status = p.main(args)

    # Take us home percy...
    sys.exit(exit_status)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import pytest

from rucio.client.metaclient import MetaClient
from rucio.common.exception import InvalidValueForKey, KeyNotFound, UnsupportedValueType, UnsupportedKeyType
from rucio.common.utils import generate_uuid as uuid
from rucio.core.meta import add_key
from rucio.db.sqla import session, models
from rucio.db.sqla.constants import DIDType, KeyType


@pytest.mark.dirty
class TestMetaClient(unittest.TestCase):

    def setUp(self):
        self.meta_client = MetaClient()

    def xtest_add_and_list_keys(self):
        """ META (CLIENTS): Add a key and List all keys."""
        key = 'key_' + str(uuid())[:20]
        ret = self.meta_client.add_key(key=key, key_type='ALL')
        assert ret
        keys = self.meta_client.list_keys()
        assert isinstance(keys, list)
        assert key in keys

    def xtest_add_and_list_values(self):
        """ META (CLIENTS): Add a value and List all values."""
        key = 'key_' + str(uuid())[:20]
        value = 'value_' + str(uuid())

        ret = self.meta_client.add_key(key=key, key_type='ALL')
        assert ret

        ret = self.meta_client.add_value(key=key, value=value)

        values = self.meta_client.list_values(key=key)
        assert isinstance(values, list)
        assert value in values

    def xtest_add_value_with_type(self):
        """ META (CLIENTS):  Add a new value to a key with a type constraint"""
        key = 'key_' + str(uuid())[:20]
        value = 'value_' + str(uuid())
        self.meta_client.add_key(key=key, key_type='ALL', value_type=str)
        self.meta_client.add_value(key=key, value=value)
        values = self.meta_client.list_values(key=key)
        assert value in values
        self.meta_client.add_value(key=key, value=1234)

    def xtest_add_value_with_regexp(self):
        """ META (CLIENTS):  Add a new value to a key with a regexp constraint"""
        key = 'guid' + str(uuid())[:20]
        value = str(uuid())
        # regexp for uuid
        regexp = '[a-f0-9]{8}[a-f0-9]{4}[a-f0-9]{4}[a-f0-9]{4}[a-f0-9]{12}'
        self.meta_client.add_key(key=key, key_type='ALL', value_regexp=regexp)
        self.meta_client.add_value(key=key, value=value)
        values = self.meta_client.list_values(key=key)
        assert value in values
        with pytest.raises(InvalidValueForKey):
            self.meta_client.add_value(key=key, value='Nimportnawak')

    def xtest_add_unsupported_type(self):
        """ META (CLIENTS):  Add an unsupported value for type """
        key = 'key_' + str(uuid())[:20]
        with pytest.raises(UnsupportedValueType):
            self.meta_client.add_key(key=key, key_type='ALL', value_type=str)

    def xtest_add_value_to_bad_key(self):
        """ META (CLIENTS):  Add a new value to a non existing key """
        value = 'value_' + str(uuid())
        with pytest.raises(KeyNotFound):
            self.meta_client.add_value(key="Nimportnawak", value=value)

    def test_add_key(self):
        """ META (CLIENTS): Add a new key """
        types = [{'type': 'FILE', 'expected': KeyType.FILE},
                 {'type': 'ALL', 'expected': KeyType.ALL},
                 {'type': 'COLLECTION', 'expected': KeyType.COLLECTION},
                 {'type': 'DATASET', 'expected': KeyType.DATASET},
                 {'type': 'D', 'expected': KeyType.DATASET},
                 {'type': 'FILE', 'expected': KeyType.FILE},
                 {'type': 'F', 'expected': KeyType.FILE},
                 {'type': 'DERIVED', 'expected': KeyType.DERIVED},
                 {'type': 'C', 'expected': KeyType.CONTAINER}]

        for key_type in types:
            key_name = 'datatype%s' % str(uuid())
            self.meta_client.add_key(key_name, key_type['type'])
            stored_key_type = session.get_session().query(models.DIDKey).filter_by(key=key_name).one()['key_type']
            assert stored_key_type, key_type['expected']

        with pytest.raises(UnsupportedKeyType):
            self.meta_client.add_key('datatype', 'A')


@pytest.mark.dirty
def test_add_key():
    """ META (CORE): Add a new key """
    types = [{'type': DIDType.FILE, 'expected': KeyType.FILE},
             {'type': DIDType.CONTAINER, 'expected': KeyType.CONTAINER},
             {'type': DIDType.DATASET, 'expected': KeyType.DATASET},
             {'type': KeyType.ALL, 'expected': KeyType.ALL},
             {'type': KeyType.DERIVED, 'expected': KeyType.DERIVED},
             {'type': KeyType.FILE, 'expected': KeyType.FILE},
             {'type': KeyType.COLLECTION, 'expected': KeyType.COLLECTION},
             {'type': KeyType.CONTAINER, 'expected': KeyType.CONTAINER},
             {'type': KeyType.DATASET, 'expected': KeyType.DATASET},
             {'type': 'FILE', 'expected': KeyType.FILE},
             {'type': 'ALL', 'expected': KeyType.ALL},
             {'type': 'COLLECTION', 'expected': KeyType.COLLECTION},
             {'type': 'DATASET', 'expected': KeyType.DATASET},
             {'type': 'D', 'expected': KeyType.DATASET},
             {'type': 'FILE', 'expected': KeyType.FILE},
             {'type': 'F', 'expected': KeyType.FILE},
             {'type': 'DERIVED', 'expected': KeyType.DERIVED},
             {'type': 'C', 'expected': KeyType.CONTAINER}]

    for key_type in types:
        key_name = 'datatype%s' % str(uuid())
        add_key(key_name, key_type['type'])
        stored_key_type = session.get_session().query(models.DIDKey).filter_by(key=key_name).one()['key_type']
        assert stored_key_type, key_type['expected']

    with pytest.raises(UnsupportedKeyType):
        add_key('datatype', DIDType.ARCHIVE)

    with pytest.raises(UnsupportedKeyType):
        add_key('datatype', 'A')



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utility functions for stomp connections
"""

import socket

from stomp import Connection

import logging


def get_stomp_brokers(brokers, port, use_ssl, vhost, reconnect_attempts, ssl_key_file, ssl_cert_file, timeout,
                      logger=logging.log):
    logger(logging.DEBUG, 'resolving broker dns alias: %s' % brokers)

    brokers_resolved = []
    for broker in brokers:
        addrinfos = socket.getaddrinfo(broker, 0, socket.AF_INET, 0, socket.IPPROTO_TCP)
        brokers_resolved.extend(ai[4][0] for ai in addrinfos)

    logger(logging.DEBUG, 'broker resolved to %s', brokers_resolved)
    conns = []
    for broker in brokers_resolved:
        if not use_ssl:
            conns.append(Connection(host_and_ports=[(broker, port)],
                                    use_ssl=False,
                                    vhost=vhost,
                                    timeout=timeout,
                                    heartbeats=(0, 1000),
                                    reconnect_attempts_max=reconnect_attempts))
        else:
            conns.append(Connection(host_and_ports=[(broker, port)],
                                    use_ssl=True,
                                    ssl_key_file=ssl_key_file,
                                    ssl_cert_file=ssl_cert_file,
                                    vhost=vhost,
                                    timeout=timeout,
                                    heartbeats=(0, 1000),
                                    reconnect_attempts_max=reconnect_attempts))
    return conns



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' add notification column to rules '''

import sqlalchemy as sa
from alembic import context, op
from alembic.op import add_column, drop_column

from rucio.db.sqla.constants import RuleNotification
from rucio.db.sqla.util import try_drop_constraint

# Alembic revision identifiers
revision = '4207be2fd914'
down_revision = '14ec5aeb64cf'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''

    if context.get_context().dialect.name in ['oracle', 'mysql']:
        add_column('rules', sa.Column('notification', sa.Enum(RuleNotification,
                                                              name='RULES_NOTIFICATION_CHK',
                                                              create_constraint=True,
                                                              values_callable=lambda obj: [e.value for e in obj]),
                                      default=RuleNotification.NO), schema=schema[:-1])
    elif context.get_context().dialect.name == 'postgresql':
        op.execute("CREATE TYPE \"RULES_NOTIFICATION_CHK\" AS ENUM('Y', 'N', 'C', 'P')")
        op.execute("ALTER TABLE %srules ADD COLUMN notification \"RULES_NOTIFICATION_CHK\"" % schema)


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''

    if context.get_context().dialect.name == 'oracle':
        try_drop_constraint('RULES_NOTIFICATION_CHK', 'rules')
        drop_column('rules', 'notification', schema=schema[:-1])

    elif context.get_context().dialect.name == 'postgresql':
        op.execute('ALTER TABLE %srules DROP CONSTRAINT IF EXISTS "RULES_NOTIFICATION_CHK", ALTER COLUMN notification TYPE CHAR' % schema)
        op.execute('ALTER TABLE %srules DROP COLUMN notification' % schema)
        op.execute('DROP TYPE \"RULES_NOTIFICATION_CHK\"')

    elif context.get_context().dialect.name == 'mysql':
        drop_column('rules', 'notification', schema=schema[:-1])




'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.29.9',
    'branch_nick': 'release-1.29-LTS',
    'revision_id': 'e63c74ba171b04b0e838475d6ade4169af7538af',
    'revno': 12094
}



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' Added vo table and vo col to rse '''

import datetime

import sqlalchemy as sa
from alembic import context
from alembic.op import (add_column, create_primary_key, create_table, create_unique_constraint,
                        drop_column, drop_constraint, drop_table, bulk_insert)


def String(*arg, **kw):
    kw['convert_unicode'] = True
    return sa.String(*arg, **kw)


# Alembic revision identifiers
revision = 'a118956323f8'
down_revision = 'd1189a09c6e0'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'postgresql', 'mysql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        # add a vo table
        vos = create_table('vos',
                           sa.Column('vo', String(3)),
                           sa.Column('description', String(255)),
                           sa.Column('email', String(255)),
                           sa.Column('created_at', sa.DateTime, default=datetime.datetime.utcnow),
                           sa.Column('updated_at', sa.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow),
                           schema=schema)
        create_primary_key('VOS_PK', 'vos', ['vo'])

        # create base vo
        bulk_insert(vos, [{'vo': 'def',
                           'description': 'Base VO',
                           'email': 'N/A'}])

        # add a vo column
        add_column('rses', sa.Column('vo', String(3), sa.ForeignKey('vos.vo', name='RSES_VOS_FK'), nullable=False, server_default='def'), schema=schema)

        # change unique constraint: (rse) -> (rse,vo)
        drop_constraint('RSES_RSE_UQ', 'rses', type_='unique', schema=schema)
        create_unique_constraint('RSES_RSE_UQ', 'rses', ['rse', 'vo'], schema=schema)


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'postgresql', 'mysql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''

        # change unique constraint: (rse, vo) -> (rse)
        drop_constraint('RSES_RSE_UQ', 'rses', type_='unique', schema=schema)
        create_unique_constraint('RSES_RSE_UQ', 'rses', ['rse'], schema=schema)

        # drop vo column
        drop_constraint('RSES_VOS_FK', 'rses', type_='foreignkey', schema=schema)
        drop_column('rses', 'vo', schema=schema)

        # drop vo table
        drop_table('vos', schema=schema)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask, Blueprint, request

from rucio.api.lock import get_dataset_locks_by_rse, get_dataset_locks, get_dataset_locks_bulk
from rucio.common.exception import RSENotFound
from rucio.common.utils import render_json
from rucio.web.rest.flaskapi.v1.common import check_accept_header_wrapper_flask, parse_scope_name, try_stream, \
    request_auth_env, response_headers, generate_http_error_flask, ErrorHandlingMethodView, json_parse


class LockByRSE(ErrorHandlingMethodView):
    """ REST APIs for dataset locks. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, rse):
        """
        ---
        summary: Get locks by rse
        description: Get all dataset locks for an associated rse.
        tags:
          - Lock
        parameters:
        - name: rse
          in: path
          description: The rse name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  did_type:
                    description: The did type to filter for.
                    type: string
                    enum: ['dataset']
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: Locks associated with the rse.
                  type: array
                  items:
                    description: A lock
                    type: object
                    properties:
                      rse_id:
                        description: The id of the associated rse.
                        type: string
                      rse:
                        description: The name of the associated rse.
                        type: string
                      scope:
                        description: The scope of the associated rse.
                        type: string
                      name:
                        description: The name of the rule.
                        type: string
                      rule_id:
                        description: The id of the rule.
                        type: string
                      account:
                        description: The associated account.
                        type: string
                      state:
                        description: The state of the rule.
                        type: string
                        enum: ['R', 'O', 'S']
                      length:
                        description: The length of the rule.
                        type: integer
                      bytes:
                        description: The bytes limit for the lock.
                        type: integere
                      accessed_at:
                        description: The last time is was accessed.
                        type: string
          401:
            description: Invalid Auth Token
          500:
            description: Wrong did type
            content:
              application/json:
                schema:
                  type: string
                  enum: ['wrong did_type specified']
          406:
            description: Not acceptable
        """
        did_type = request.args.get('did_type', default=None)
        if did_type != 'dataset':
            return 'Wrong did_type specified', 500

        try:
            def generate(vo):
                for lock in get_dataset_locks_by_rse(rse, vo=vo):
                    yield render_json(**lock) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')))
        except RSENotFound as error:
            return generate_http_error_flask(404, error)


class LocksByScopeName(ErrorHandlingMethodView):
    """ REST APIs for dataset locks. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, scope_name):
        """
        ---
        summary: Get locks by scope
        description: Get all dataset locks for an associated rse.
        tags:
          - Lock
        parameters:
        - name: scope_name
          in: path
          description: The scope name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  did_type:
                    description: The did type to filter for.
                    type: string
                    enum: ['dataset']
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: Locks associated with the rse.
                  type: array
                  items:
                    description: A lock
                    type: object
                    properties:
                      rse_id:
                        description: The id of the associated rse.
                        type: string
                      rse:
                        description: The name of the associated rse.
                        type: string
                      scope:
                        description: The scope of the associated rse.
                        type: string
                      name:
                        description: The name of the rule.
                        type: string
                      rule_id:
                        description: The id of the rule.
                        type: string
                      account:
                        description: The associated account.
                        type: string
                      state:
                        description: The state of the rule.
                        type: string
                        enum: ['R', 'O', 'S']
                      length:
                        description: The length of the rule.
                        type: integer
                      bytes:
                        description: The bytes limit for the lock.
                        type: integere
                      accessed_at:
                        description: The last time is was accessed.
                        type: string
          401:
            description: Invalid Auth Token
          500:
            description: Wrong did type
            content:
              application/json:
                schema:
                  type: string
                  enum: ['wrong did_type specified']
          406:
            description: Not acceptable
        """
        did_type = request.args.get('did_type', default=None)
        if did_type != 'dataset':
            return 'Wrong did_type specified', 500

        try:
            scope, name = parse_scope_name(scope_name, request.environ.get('vo'))

            def generate(vo):
                for lock in get_dataset_locks(scope, name, vo=vo):
                    yield render_json(**lock) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')))
        except ValueError as error:
            return generate_http_error_flask(400, error)


class DatasetLocksForDids(ErrorHandlingMethodView):
    """ REST APIs for multiple dataset locks. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def post(self):
        """ get locks for a given scope, name.

        :resheader Content-Type: application/x-json-stream
        :status 200: OK.
        :status 400: Wrong DID type.
        :returns: Line separated list of dictionary with lock information.
        ---
        summary: Get locks by dids
        description: Get all dataset locks for the associated dids.
        tags:
          - Lock
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  dids:
                    description: The dids associated with the locks.
                    type: array
                    items:
                      type: object
                      description: A did
                      required:
                        - scope
                        - name
                      properties:
                        scope:
                          description: The scope of the did.
                          type: string
                        name:
                          description: The name of the did.
                          type: string
                        type:
                          description: The type of the did.
                          type: string
                          enum: ['dataset', 'container']
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: Locks associated with the rse.
                  type: array
                  items:
                    description: A lock
                    type: object
                    properties:
                      rse_id:
                        description: The id of the associated rse.
                        type: string
                      rse:
                        description: The name of the associated rse.
                        type: string
                      scope:
                        description: The scope of the associated rse.
                        type: string
                      name:
                        description: The name of the rule.
                        type: string
                      rule_id:
                        description: The id of the rule.
                        type: string
                      account:
                        description: The associated account.
                        type: string
                      state:
                        description: The state of the rule.
                        type: string
                        enum: ['R', 'O', 'S']
                      length:
                        description: The length of the rule.
                        type: integer
                      bytes:
                        description: The bytes limit for the lock.
                        type: integere
                      accessed_at:
                        description: The last time is was accessed.
                        type: string
          401:
            description: Invalid Auth Token
          400:
            description: Wrong did type
            content:
              application/json:
                schema:
                  type: string
                  enum: ['Can not find the list of DIDs in the data. Use "dids" keyword.']
          406:
            description: Not acceptable

        """

        data = json_parse(types=(dict,))
        try:
            dids = data["dids"]
        except KeyError:
            return 'Can not find the list of DIDs in the data. Use "dids" keyword.', 400
        vo = request.environ.get('vo')
        try:
            locks = get_dataset_locks_bulk(dids, vo)        # removes duplicates

            def generate(locks):
                for lock in locks:
                    lock["scope"] = str(lock["scope"])
                    yield render_json(**lock) + '\n'
            return try_stream(generate(locks))

        except ValueError as error:
            return generate_http_error_flask(400, error)


def blueprint():
    bp = Blueprint('locks', __name__, url_prefix='/locks')

    lock_by_rse_view = LockByRSE.as_view('lock_by_rse')
    bp.add_url_rule('/<rse>', view_func=lock_by_rse_view, methods=['get', ])

    lock_by_scope_name_view = LocksByScopeName.as_view('locks_by_scope_name')
    bp.add_url_rule('/<path:scope_name>', view_func=lock_by_scope_name_view, methods=['get', ])

    locks_for_dids_view = DatasetLocksForDids.as_view('locks_for_dids')
    bp.add_url_rule('/bulk_locks_for_dids', view_func=locks_for_dids_view, methods=['post', ])

    bp.before_request(request_auth_env)
    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import rucio.api.permission
import rucio.common.exception

from rucio.core import scope as core_scope
from rucio.common.types import InternalAccount, InternalScope
from rucio.common.schema import validate_schema
from rucio.db.sqla.session import read_session, transactional_session


@read_session
def list_scopes(filter_={}, vo='def', session=None):
    """
    Lists all scopes.

    :param filter_: Dictionary of attributes by which the input data should be filtered
    :param vo: The VO to act on.
    :param session: The database session in use.

    :returns: A list containing all scopes.
    """
    # If filter is empty, create a new dict to avoid overwriting the function's default
    if not filter_:
        filter_ = {}

    if 'scope' in filter_:
        filter_['scope'] = InternalScope(scope=filter_['scope'], vo=vo)
    else:
        filter_['scope'] = InternalScope(scope='*', vo=vo)
    return [scope.external for scope in core_scope.list_scopes(filter_=filter_, session=session)]


@transactional_session
def add_scope(scope, account, issuer, vo='def', session=None):
    """
    Creates a scope for an account.

    :param account: The account name.
    :param scope: The scope identifier.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    validate_schema(name='scope', obj=scope, vo=vo)

    kwargs = {'scope': scope, 'account': account}
    if not rucio.api.permission.has_permission(issuer=issuer, vo=vo, action='add_scope', kwargs=kwargs, session=session):
        raise rucio.common.exception.AccessDenied('Account %s can not add scope' % (issuer))

    scope = InternalScope(scope, vo=vo)
    account = InternalAccount(account, vo=vo)

    core_scope.add_scope(scope, account, session=session)


@read_session
def get_scopes(account, vo='def', session=None):
    """
    Gets a list of all scopes for an account.

    :param account: The account name.
    :param vo: The VO to act on.
    :param session: The database session in use.

    :returns: A list containing the names of all scopes for this account.
    """

    account = InternalAccount(account, vo=vo)

    return [scope.external for scope in core_scope.get_scopes(account, session=session)]



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import pytest

import rucio.core.config as config_core
from rucio.client.configclient import ConfigClient
from rucio.common import exception
from rucio.common.utils import generate_uuid


class TestConfigCore:

    def test_get_config_sections(self):
        """ CONFIG (CORE): Retreive configuration section only """
        expected_sections = [str(generate_uuid()), str(generate_uuid())]
        for section in expected_sections:
            config_core.set(section, str(generate_uuid()), str(generate_uuid()))
        sections = config_core.sections(use_cache=False)
        for section in expected_sections:
            assert section in sections

    def test_get_and_set_section_option(self):
        """ CONFIG (CORE): Retreive configuration option only """
        # get and set
        section = str(generate_uuid())
        option = str(generate_uuid())
        expected_value = str(generate_uuid())
        config_core.set(section=section, option=option, value=expected_value)
        value = config_core.get(section, option, use_cache=False)
        assert value == expected_value

        # default value
        section = str(generate_uuid())
        config_core.set(section=section, option=str(generate_uuid()), value=str(generate_uuid()))
        default_value = 'default'
        value = config_core.get(section, 'new_option', default=default_value, use_cache=False)
        assert value == default_value

        # key with space character
        section = str(generate_uuid() + ' ')
        option = str(generate_uuid() + ' ')
        expected_value = str(generate_uuid())
        config_core.set(section=section, option=option, value=expected_value)
        value = config_core.get(section, option, use_cache=False)
        assert value == expected_value


def test_config_section_contextless():
    config = ConfigClient()
    test_section_1 = generate_uuid()
    test_section_2 = generate_uuid()
    config.set_config_option(test_section_1, 'a', 'b')
    config.set_config_option(test_section_2, 'c', 'd')

    value = config.get_config(None, None)

    assert isinstance(value, dict)
    assert test_section_1 in value.keys()
    assert test_section_2 in value.keys()


class TestConfigClients(unittest.TestCase):

    def setUp(self):
        self.c = ConfigClient()
        self.test_section_1 = str(generate_uuid())
        self.test_section_2 = str(generate_uuid())
        self.test_option_s = 'string'
        self.test_option_b = 'bool'
        self.test_option_i = 'int'
        self.test_option_f = 'float'
        self.test_option_sv = 'iddqd'
        self.test_option_bv = 'True'
        self.test_option_iv = '543210'
        self.test_option_fv = '3.1415'
        self.c.set_config_option(self.test_section_1, self.test_option_s, self.test_option_sv)
        self.c.set_config_option(self.test_section_1, self.test_option_b, self.test_option_bv)
        self.c.set_config_option(self.test_section_2, self.test_option_i, self.test_option_iv)
        self.c.set_config_option(self.test_section_2, self.test_option_f, self.test_option_fv)

    def tearDown(self):
        self.c = None

    def test_get_config_all(self):
        """ CONFIG (CLIENT): Retrieve configuration values and check for correctness """
        tmp = self.c.get_config(None, None)
        assert isinstance(tmp, dict)
        assert self.test_section_1 in tmp.keys()
        assert self.test_option_s in tmp[self.test_section_1]
        assert self.test_option_b in tmp[self.test_section_1]
        assert self.test_option_sv == tmp[self.test_section_1][self.test_option_s]
        assert tmp[self.test_section_1][self.test_option_b]
        assert self.test_option_i in tmp[self.test_section_2]
        assert self.test_option_f in tmp[self.test_section_2]
        assert 543210 == tmp[self.test_section_2][self.test_option_i]
        assert 3.1415 == tmp[self.test_section_2][self.test_option_f]

    def test_get_config_section(self):
        """ CONFIG (CLIENT): Retrieve configuration section only """
        tmp = self.c.get_config(self.test_section_1, None)
        assert isinstance(tmp, dict)
        assert self.test_option_s in tmp.keys()
        assert self.test_option_b in tmp.keys()

    def test_get_config_section_option(self):
        """ CONFIG (CLIENT): Retrieve configuration option only """
        tmp = self.c.get_config(self.test_section_1, self.test_option_s)
        assert tmp == self.test_option_sv

        with pytest.raises(exception.ConfigNotFound):
            self.c.get_config(self.test_section_1, 'no_option')

    def test_set_and_get_config_value_special_strings(self):
        for test_option_description, option_value in [
            ('dot', '.'),
            ('slash', '/'),
            ('aPath', 'a/b/c/../'),
            ('percentEncodedSpecialChar', '%2E'),
            ('urlParameters', 'a?x=y'),
        ]:
            self.c.set_config_option(self.test_section_1, test_option_description, option_value)
            retrieved_value = self.c.get_config(self.test_section_1, test_option_description)
            assert retrieved_value == option_value

    def test_set_config_option_via_deprecated_url(self):
        """
        The format of the /config endpoint was recently changed, but we still support the old
        format for API calls for the transition period.
        TODO: remove this test
        """
        self.c.set_config_option(self.test_section_1, self.test_option_s + 'ViaUrl', self.test_option_sv, use_body_for_params=False)
        self.c.set_config_option(self.test_section_1, self.test_option_b + 'ViaUrl', self.test_option_bv, use_body_for_params=False)
        self.c.set_config_option(self.test_section_2, self.test_option_i + 'ViaUrl', self.test_option_iv, use_body_for_params=False)
        self.c.set_config_option(self.test_section_2, self.test_option_f + 'ViaUrl', self.test_option_fv, use_body_for_params=False)
        tmp = self.c.get_config(None, None)
        assert self.test_option_s + 'ViaUrl' in tmp[self.test_section_1]
        assert self.test_option_b + 'ViaUrl' in tmp[self.test_section_1]
        assert self.test_option_i + 'ViaUrl' in tmp[self.test_section_2]
        assert self.test_option_f + 'ViaUrl' in tmp[self.test_section_2]



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' new bad_pfns table and bad_replicas changes '''

import datetime

import sqlalchemy as sa
from alembic import context, op
from alembic.op import (create_primary_key, create_check_constraint,
                        drop_constraint, create_table, create_foreign_key,
                        add_column, create_index, drop_table, drop_column,
                        drop_index)

from rucio.db.sqla.constants import BadPFNStatus
from rucio.db.sqla.util import try_drop_constraint

# Alembic revision identifiers
revision = 'b96a1c7e1cc4'
down_revision = '1f46c5f240ac'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''

    if context.get_context().dialect.name in ['oracle', 'postgresql']:
        # Create new bad_pfns table
        create_table('bad_pfns',
                     sa.Column('path', sa.String(2048)),
                     sa.Column('state', sa.Enum(BadPFNStatus,
                                                name='BAD_PFNS_STATE_CHK',
                                                create_constraint=True,
                                                values_callable=lambda obj: [e.value for e in obj]),
                               default=BadPFNStatus.SUSPICIOUS),
                     sa.Column('reason', sa.String(255)),
                     sa.Column('account', sa.String(25)),
                     sa.Column('expires_at', sa.DateTime),
                     sa.Column('created_at', sa.DateTime, default=datetime.datetime.utcnow),
                     sa.Column('updated_at', sa.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow))

        create_primary_key('BAD_PFNS_PK', 'bad_pfns', ['path', 'state'])
        create_foreign_key('BAD_PFNS_ACCOUNT_FK', 'bad_pfns', 'accounts', ['account'], ['account'])

        try_drop_constraint('BAD_REPLICAS_STATE_CHK', 'bad_replicas')
        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S', 'T')")

        # Add new column to bad_replicas table
        add_column('bad_replicas', sa.Column('expires_at', sa.DateTime()), schema=schema[:-1])

        # Change PK
        drop_constraint('BAD_REPLICAS_STATE_PK', 'bad_replicas', type_='primary')
        create_primary_key('BAD_REPLICAS_STATE_PK', 'bad_replicas', ['scope', 'name', 'rse_id', 'state', 'created_at'])

        # Add new Index to Table
        create_index('BAD_REPLICAS_EXPIRES_AT_IDX', 'bad_replicas', ['expires_at'])

    elif context.get_context().dialect.name == 'mysql':
        # Create new bad_pfns table
        create_table('bad_pfns',
                     sa.Column('path', sa.String(2048)),
                     sa.Column('state', sa.Enum(BadPFNStatus,
                                                name='BAD_PFNS_STATE_CHK',
                                                create_constraint=True,
                                                values_callable=lambda obj: [e.value for e in obj]),
                               default=BadPFNStatus.SUSPICIOUS),
                     sa.Column('reason', sa.String(255)),
                     sa.Column('account', sa.String(25)),
                     sa.Column('expires_at', sa.DateTime),
                     sa.Column('created_at', sa.DateTime, default=datetime.datetime.utcnow),
                     sa.Column('updated_at', sa.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow))

        create_primary_key('BAD_PFNS_PK', 'bad_pfns', ['path', 'state'])
        create_foreign_key('BAD_PFNS_ACCOUNT_FK', 'bad_pfns', 'accounts', ['account'], ['account'])

        op.execute('ALTER TABLE ' + schema + 'bad_replicas DROP CHECK BAD_REPLICAS_STATE_CHK')  # pylint: disable=no-member
        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S', 'T')")

        # Add new column to bad_replicas table
        add_column('bad_replicas', sa.Column('expires_at', sa.DateTime()), schema=schema[:-1])

        # Change PK
        drop_constraint('BAD_REPLICAS_STATE_PK', 'bad_replicas', type_='primary')
        create_primary_key('BAD_REPLICAS_STATE_PK', 'bad_replicas', ['scope', 'name', 'rse_id', 'state', 'created_at'])

        # Add new Index to Table
        create_index('BAD_REPLICAS_EXPIRES_AT_IDX', 'bad_replicas', ['expires_at'])


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''

    if context.get_context().dialect.name == 'oracle':
        drop_table('bad_pfns')
        drop_index('BAD_REPLICAS_EXPIRES_AT_IDX', 'bad_replicas')

        try_drop_constraint('BAD_REPLICAS_STATE_CHK', 'bad_replicas')
        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S')")

        drop_column('bad_replicas', 'expires_at')
        drop_constraint('BAD_REPLICAS_STATE_PK', 'bad_replicas', type_='primary')
        create_primary_key('BAD_REPLICAS_STATE_PK', 'bad_replicas', ['scope', 'name', 'rse_id', 'created_at'])

    elif context.get_context().dialect.name == 'postgresql':
        drop_table('bad_pfns')
        drop_index('BAD_REPLICAS_EXPIRES_AT_IDX', 'bad_replicas')

        op.execute('ALTER TABLE ' + schema + 'bad_replicas DROP CONSTRAINT IF EXISTS "BAD_REPLICAS_STATE_CHK", ALTER COLUMN state TYPE CHAR')  # pylint: disable=no-member
        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S')")

        drop_column('bad_replicas', 'expires_at', schema=schema[:-1])
        drop_constraint('BAD_REPLICAS_STATE_PK', 'bad_replicas', type_='primary')
        create_primary_key('BAD_REPLICAS_STATE_PK', 'bad_replicas', ['scope', 'name', 'rse_id', 'created_at'])

    elif context.get_context().dialect.name == 'mysql':
        drop_table('bad_pfns')
        drop_index('BAD_REPLICAS_EXPIRES_AT_IDX', 'bad_replicas')

        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S')")

        drop_column('bad_replicas', 'expires_at', schema=schema[:-1])
        drop_constraint('BAD_REPLICAS_STATE_PK', 'bad_replicas', type_='primary')
        create_primary_key('BAD_REPLICAS_STATE_PK', 'bad_replicas', ['scope', 'name', 'rse_id', 'created_at'])



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta

import pytest
from sqlalchemy import delete

from rucio.common.utils import generate_uuid
from rucio.core.did import attach_dids, add_did
from rucio.core.distance import add_distance
from rucio.core.replica import add_replica
from rucio.core.request import queue_requests, get_request_by_did, release_waiting_requests_per_deadline, release_all_waiting_requests, release_waiting_requests_fifo, release_waiting_requests_grouped_fifo, release_waiting_requests_per_free_volume
from rucio.core.rse import set_rse_transfer_limits
from rucio.daemons.conveyor import throttler, preparer
from rucio.db.sqla import models
from rucio.db.sqla.constants import DIDType, RequestType, RequestState
from rucio.tests.common import skiplimitedsql


@pytest.fixture
def connected_rse_pair(vo, rse_factory, db_session):
    source_rse, source_rse_id = rse_factory.make_mock_rse()
    dest_rse, dest_rse_id = rse_factory.make_mock_rse()
    add_distance(source_rse_id, dest_rse_id, ranking=10)

    return source_rse, source_rse_id, dest_rse, dest_rse_id


def _delete_requests(scope, names, ids=None, session=None):
    session.execute(
        delete(
            models.Request
        ).where(
            models.Request.scope == scope,
            models.Request.name.in_(names)
        )
    )
    if ids:
        session.execute(
            delete(
                models.Request
            ).where(
                models.Request.id.in_(ids)
            )
        )
    session.commit()


def _add_test_replicas_and_request(request_configs, scope, account, session):
    """
    Generate replicas and associated requests; one for each config in request_configs.
    """
    names = []
    requests = []
    for config in request_configs:
        name = generate_uuid()
        request = {
            'source_rse_id': '',
            'dest_rse_id': '',
            'request_type': RequestType.TRANSFER,
            'request_id': generate_uuid(),
            'name': name,
            'bytes': 1,
            'scope': scope,
            'retry_count': 1,
            'rule_id': generate_uuid(),
            'requested_at': datetime.now(),
            'account': account,
            'attributes': {
                'activity': 'User Subscription',
                'bytes': 1,
                'md5': '',
                'adler32': '',
            },
        }
        request.update({k: w for k, w in config.items() if k != 'attributes'})
        if config.get('attributes'):
            request['attributes'].update({k: w for k, w in config['attributes'].items()})

        names.append(name)
        requests.append(request)

        add_replica(request['source_rse_id'], request['scope'], request['name'], request['bytes'], account, session=session)

    queue_requests(requests, session=session)
    session.commit()
    return names


@pytest.mark.noparallel(reason='uses preparer and throttler')
@pytest.mark.usefixtures("core_config_mock", "file_config_mock")
@pytest.mark.parametrize("file_config_mock", [{"overrides": [
    ('conveyor', 'use_preparer', 'true')
]}], indirect=True)
class TestSimpleLimits:
    """
    Test the behavior of throttler on simple cases without overlapping limits.
    """

    user_activity = 'User Subscription'
    user_activity2 = 'User Subscription2'
    all_activities = 'all_activities'

    def _add_replicas_and_request(self, source_rse_id, dest_rse_id, scope, account, session):
        return _add_test_replicas_and_request(
            scope=scope, account=account, session=session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2000),
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2021),  # requested after the request below but small enough for max_volume check
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'bytes': 3000},
                },
            ]
        )

    def test_dest_all_act_grouped_fifo_all(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release all waiting requests (DEST - ALL ACT - GFIFO). """
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # no threshold when releasing -> release all waiting requests
        set_rse_transfer_limits(dest_rse_id, max_transfers=1, activity=self.all_activities, strategy='grouped_fifo', session=db_session)

        name1, name2, name3, name4 = self._add_replicas_and_request(source_rse_id, dest_rse_id, scope=mock_scope, account=root_account, session=db_session)
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()

        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request_1['state'] == RequestState.WAITING
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request_2['state'] == RequestState.WAITING
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id)
        assert request_4['state'] == RequestState.WAITING

        db_session.execute(
            delete(
                models.RSETransferLimit
            ).where(
                models.RSETransferLimit.rse_id.in_([source_rse_id, dest_rse_id])
            )
        )
        db_session.commit()

        throttler.run_once(logger=print, session=db_session)
        db_session.commit()

        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request_1['state'] == RequestState.QUEUED
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request_2['state'] == RequestState.QUEUED
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request_3['state'] == RequestState.QUEUED
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id)
        assert request_4['state'] == RequestState.QUEUED

    def test_dest_all_act_grouped_fifo_nothing(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release nothing (DEST - ALL ACT - GFIFO). """
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # four waiting requests and one active requests but threshold is 1
        # more than 80% of the transfer limit are already used -> release nothing
        set_rse_transfer_limits(dest_rse_id, max_transfers=1, activity=self.all_activities, strategy='grouped_fifo', session=db_session)
        request = models.Request(dest_rse_id=dest_rse_id, bytes=2, activity=self.user_activity, state=RequestState.SUBMITTED)
        request.save(session=db_session)
        name1, name2, name3, name4 = self._add_replicas_and_request(source_rse_id, dest_rse_id, scope=mock_scope, account=root_account, session=db_session)
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request_1['state'] == RequestState.WAITING
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request_2['state'] == RequestState.WAITING
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id)
        assert request_4['state'] == RequestState.WAITING

    @skiplimitedsql
    def test_dest_all_act_grouped_fifo_subset(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (DEST - ALL ACT - GFIFO). """
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=10, max_transfers=1, deadline=1, strategy='grouped_fifo', session=db_session)
        name1, name2, name3, name4 = self._add_replicas_and_request(source_rse_id, dest_rse_id, scope=mock_scope, account=root_account, session=db_session)
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        # released because it got requested first
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request_1['state'] == RequestState.QUEUED
        # released because the DID is attached to the same dataset
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request_2['state'] == RequestState.QUEUED
        # released because of available volume
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request_3['state'] == RequestState.QUEUED
        # still waiting because there is no free volume
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id)
        assert request_4['state'] == RequestState.WAITING
        # deadline check should not work for destination RSEs - only for reading

    def test_dest_per_act_fifo_release_all(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release all waiting requests (DEST - ACT - FIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # all activities limit applies to each activity -> release only one transfer
        set_rse_transfer_limits(dest_rse_id, max_transfers=1, activity=self.all_activities, strategy='fifo', session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING

        # active transfers + waiting requests are less than the threshold -> release all waiting requests
        _delete_requests(mock_scope, [name1, name2], session=db_session)
        set_rse_transfer_limits(dest_rse_id, activity=self.user_activity, max_transfers=3, strategy='fifo', session=db_session)
        set_rse_transfer_limits(dest_rse_id, max_transfers=3, activity=self.all_activities, strategy='fifo', session=db_session)
        request = models.Request(dest_rse_id=dest_rse_id, activity=self.user_activity, state=RequestState.SUBMITTED)
        request.save(session=db_session)
        name1, = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED

    def test_dest_per_act_fifo_release_nothing(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release nothing (DEST - ACT - FIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # two waiting requests and one active requests but threshold is 1
        # more than 80% of the transfer limit are already used -> release nothing
        set_rse_transfer_limits(dest_rse_id, max_transfers=1, activity=self.user_activity, strategy='fifo', session=db_session)
        request = models.Request(dest_rse_id=dest_rse_id, bytes=2, activity=self.user_activity, state=RequestState.SUBMITTED)
        request.save(session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.WAITING
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING

    def test_dest_per_act_fifo_release_subset(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (DEST - ACT - FIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # two waiting requests and no active requests but threshold is 1 -> release only 1 request
        set_rse_transfer_limits(dest_rse_id, activity=self.user_activity, max_transfers=1, strategy='fifo', session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        db_session.commit()
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING

    def test_source_per_act_fifo_release_subset(self, rse_factory, mock_scope, vo, root_account, db_session):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (SRC - ACT - FIFO). """

        if db_session.bind.dialect.name == 'mysql':
            return True

        _, source_rse_id = rse_factory.make_mock_rse()
        _, dest_rse_id = rse_factory.make_mock_rse()
        _, dest_rse_id2 = rse_factory.make_mock_rse()
        _, dest_rse_id3 = rse_factory.make_mock_rse()
        for rse_id in (dest_rse_id, dest_rse_id2, dest_rse_id3):
            add_distance(source_rse_id, rse_id, ranking=10)
        # two waiting requests and no active requests but threshold is 1 for one activity
        # one waiting request and no active requests but threshold is 0 for other activity -> release only 1 request for one activity
        set_rse_transfer_limits(source_rse_id, activity=self.user_activity, max_transfers=1, strategy='fifo', direction='source')
        set_rse_transfer_limits(source_rse_id, activity=self.user_activity2, max_transfers=0, strategy='fifo', direction='source')
        name1, name2, name3 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2018),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id2,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id3,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity2},
                },
            ]
        )

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id2)
        assert request2['state'] == RequestState.WAITING
        request3 = get_request_by_did(mock_scope, name3, dest_rse_id3)
        assert request3['state'] == RequestState.WAITING

    def test_source_all_act_fifo_release_subset(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (SRC - ALL ACT - FIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # two waiting requests and no active requests but threshold is 1 -> release only 1 request
        set_rse_transfer_limits(source_rse_id, activity=self.all_activities, max_transfers=1, strategy='fifo', direction='source', session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING

    def test_dest_all_act_fifo_release_subset(self, rse_factory, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (DEST - ALL ACT - FIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        _, source_rse_id2 = rse_factory.make_mock_rse()
        _, dest_rse_id2 = rse_factory.make_mock_rse()
        add_distance(source_rse_id, dest_rse_id2, ranking=10)
        add_distance(source_rse_id2, dest_rse_id, ranking=10)
        add_distance(source_rse_id2, dest_rse_id2, ranking=10)

        # two waiting requests and no active requests but threshold 1 for one dest RSE
        # one waiting request and no active requests but threshold 0 for another dest RSE -> release only 1 request on one dest RSE
        set_rse_transfer_limits(dest_rse_id, activity=self.all_activities, max_transfers=1, strategy='fifo', session=db_session)
        set_rse_transfer_limits(dest_rse_id2, activity=self.all_activities, max_transfers=0, strategy='fifo', session=db_session)
        name1, name2, name3 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2018),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id2,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity2},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id2,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity2},
                },
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        # release because max_transfers=1
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        # waiting because limit already exceeded
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING
        request3 = get_request_by_did(mock_scope, name3, dest_rse_id2)
        assert request3['state'] == RequestState.WAITING

    @skiplimitedsql
    def test_source_all_act_grouped_fifo_subset(self, rse_factory, mock_scope, db_session, root_account, connected_rse_pair):
        """ THROTTLER (CLIENTS): throttler release subset of waiting requests (SRC - ALL ACT - GFIFO). """
        if db_session.bind.dialect.name == 'mysql':
            return True

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        _, dest_rse_id2 = rse_factory.make_mock_rse()
        add_distance(source_rse_id, dest_rse_id2, ranking=10)

        set_rse_transfer_limits(source_rse_id, self.all_activities, volume=10, max_transfers=1, deadline=0, direction='source', strategy='grouped_fifo', session=db_session)
        name1, name2, name3, name4 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2000),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id2,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.all_activities},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2021),
                    'attributes': {'activity': self.all_activities},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id2,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.all_activities},
                },
            ]
        )
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request_1['state'] == RequestState.WAITING
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id2, session=db_session)
        assert request_2['state'] == RequestState.WAITING
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id2, session=db_session)
        assert request_4['state'] == RequestState.WAITING

        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        # released because it got requested first
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request_1['state'] == RequestState.QUEUED
        # released because the DID is attached to the same dataset
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id2)
        assert request_2['state'] == RequestState.QUEUED
        # still waiting, volume check is only working for destination RSEs (writing)
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id2)
        assert request_4['state'] == RequestState.WAITING


@pytest.mark.noparallel(reason='uses preparer and throttler')
@pytest.mark.usefixtures("core_config_mock", "file_config_mock")
@pytest.mark.parametrize("file_config_mock", [{"overrides": [
    ('conveyor', 'use_preparer', 'true')
]}], indirect=True)
class TestOverlappingLimits:
    user_activity = 'User Subscription'
    user_activity2 = 'User Subscription2'
    user_activity3 = 'User Subscription3'
    all_activities = 'all_activities'

    def test_source_dest_mixed_act(self, rse_factory, mock_scope, vo, root_account, db_session):
        """
        """

        if db_session.bind.dialect.name == 'mysql':
            return True

        _, source_rse_id = rse_factory.make_mock_rse()
        _, dest_rse_id = rse_factory.make_mock_rse()
        _, dest_rse_id2 = rse_factory.make_mock_rse()
        _, dest_rse_id3 = rse_factory.make_mock_rse()
        for rse_id in (dest_rse_id, dest_rse_id2, dest_rse_id3):
            add_distance(source_rse_id, rse_id, ranking=10)
        # two waiting requests and no active requests but threshold is 1 for one activity
        # one waiting request and no active requests but threshold is 0 for other activity -> release only 1 request for one activity
        set_rse_transfer_limits(source_rse_id, activity=self.user_activity, max_transfers=1, strategy='fifo', direction='source')
        set_rse_transfer_limits(source_rse_id, activity=self.all_activities, max_transfers=3, strategy='fifo', direction='source')
        set_rse_transfer_limits(dest_rse_id, activity=self.user_activity2, max_transfers=1, strategy='fifo', direction='destination')
        set_rse_transfer_limits(dest_rse_id, activity=self.all_activities, max_transfers=4, strategy='fifo', direction='destination')
        name1, name2, name3, name4, name5 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2018),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2019),
                    'attributes': {'activity': self.user_activity2},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity2},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': self.user_activity3},
                }
            ]
        )

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        throttler.run_once(logger=print, session=db_session)
        db_session.commit()
        request = get_request_by_did(mock_scope, name1, dest_rse_id)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id)
        assert request2['state'] == RequestState.WAITING
        request3 = get_request_by_did(mock_scope, name3, dest_rse_id)
        assert request3['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name4, dest_rse_id)
        assert request['state'] == RequestState.WAITING
        request2 = get_request_by_did(mock_scope, name5, dest_rse_id)
        assert request2['state'] == RequestState.QUEUED


@pytest.mark.noparallel(reason='uses preparer and throttler')
@pytest.mark.usefixtures("core_config_mock", "file_config_mock")
@pytest.mark.parametrize("file_config_mock", [{"overrides": [
    ('conveyor', 'use_preparer', 'true')
]}], indirect=True)
class TestRequestCoreRelease:
    """Test release methods used in throttler."""
    user_activity = 'User Subscription'
    user_activity2 = 'User Subscription2'
    all_activities = 'all_activities'

    @skiplimitedsql
    def test_release_waiting_requests_per_free_volume(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ REQUEST (CORE): release waiting requests that fit grouped in available volume."""
        # release unattached requests that fit in available volume with respect to already submitted transfers

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        dummy_request = models.Request(dest_rse_id=dest_rse_id, bytes=2, activity=self.all_activities, state=RequestState.SUBMITTED)
        dummy_request.save(session=db_session)
        volume = 10
        set_rse_transfer_limits(dest_rse_id, self.user_activity, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=volume, max_transfers=1, session=db_session)
        name1, name2, name3 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2015),
                    'attributes': {'bytes': 8},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'bytes': 2},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2000),
                    'attributes': {'bytes': 10},
                },
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_per_free_volume(dest_rse_id, volume=volume, session=db_session)
        # released because small enough
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        # still waiting because requested later and to big
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING
        # still waiting because too big
        request = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING

        # release attached requests that fit together with the dataset in available volume with respect to already submitted transfers
        _delete_requests(mock_scope, [name1, name2, name3], ids=[dummy_request['id']], session=db_session)
        dummy_request = models.Request(dest_rse_id=dest_rse_id, bytes=2, activity=self.all_activities, state=RequestState.SUBMITTED)
        dummy_request.save(session=db_session)
        volume = 10
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=volume, max_transfers=1, session=db_session)
        name1, name2, name3, name4 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2015),
                    'attributes': {'bytes': 6},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'bytes': 2},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2000),
                    'attributes': {'bytes': 10},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2030),
                    'attributes': {'bytes': 2},
                },
            ]
        )
        dataset1_name = generate_uuid()
        add_did(mock_scope, dataset1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset1_name, [{'name': name1, 'scope': mock_scope}, {'name': name4, 'scope': mock_scope}], root_account, session=db_session)
        dataset2_name = generate_uuid()
        add_did(mock_scope, dataset2_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset2_name, [{'name': name2, 'scope': mock_scope}, {'name': name3, 'scope': mock_scope}], root_account, session=db_session)
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_per_free_volume(dest_rse_id, volume=volume, session=db_session)
        # released because dataset fits in volume
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name4, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        # waiting because dataset is too big
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING
        request = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING

        # release requests with no available volume -> release nothing
        _delete_requests(mock_scope, [name1, name2, name3, name4], ids=[dummy_request['id']], session=db_session)
        volume = 0
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=volume, max_transfers=1, session=db_session)
        name1, = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2015)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_per_free_volume(dest_rse_id, volume=volume, session=db_session)
        # waiting because no available volume
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING

    @skiplimitedsql
    def test_release_waiting_requests_grouped_fifo(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ REQUEST (CORE): release waiting requests based on grouped FIFO. """

        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # set volume and deadline to 0 to check first without releasing extra requests
        set_rse_transfer_limits(dest_rse_id, self.user_activity, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=0, max_transfers=1, session=db_session)

        # one request with an unattached DID -> one request should be released
        name, = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2015)},
            ]
        )

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_grouped_fifo(dest_rse_id, count=1, volume=0, deadline=0, session=db_session)
        request = get_request_by_did(mock_scope, name, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED

        # one request with an attached DID -> one request should be released
        _delete_requests(mock_scope, [name], session=db_session)
        name, = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2015)},
            ]
        )
        dataset_name = generate_uuid()
        add_did(mock_scope, dataset_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_name, [{'name': name, 'scope': mock_scope}], root_account, session=db_session)
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_grouped_fifo(dest_rse_id, count=1, volume=0, deadline=0, session=db_session)
        request = get_request_by_did(mock_scope, name, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED

        # five requests with different requested_at and multiple attachments per collection -> release only one request -> two requests of one collection should be released
        _delete_requests(mock_scope, [name], session=db_session)
        name1, name2, name3, name4, name5 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2000)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2015)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2010)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
            ]
        )
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        dataset_2_name = generate_uuid()
        add_did(mock_scope, dataset_2_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}, {'name': name2, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_2_name, [{'name': name3, 'scope': mock_scope}, {'name': name4, 'scope': mock_scope}], root_account, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_grouped_fifo(dest_rse_id, count=1, deadline=0, volume=0, session=db_session)
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request_1['state'] == RequestState.QUEUED
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request_2['state'] == RequestState.QUEUED
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id, session=db_session)
        assert request_4['state'] == RequestState.WAITING
        request_5 = get_request_by_did(mock_scope, name5, dest_rse_id, session=db_session)
        assert request_5['state'] == RequestState.WAITING

        # with maximal volume check -> release one request -> three requests should be released because of attachments and free volume space
        _delete_requests(mock_scope, [name1, name2, name3, name4, name5], session=db_session)
        name1, name2, name3, name4 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2000)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
                # 2021: requested after the request below but small enough for max_volume check
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2021)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020), 'attributes': {'bytes': 3000}},
            ]
        )
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=10, max_transfers=1, session=db_session)

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        amount_updated_requests = release_waiting_requests_grouped_fifo(dest_rse_id, count=1, deadline=0, volume=10, session=db_session)
        assert amount_updated_requests == 3
        # released because it got requested first
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request_1['state'] == RequestState.QUEUED
        # released because the DID is attached to the same dataset
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request_2['state'] == RequestState.QUEUED
        # released because of available volume
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request_3['state'] == RequestState.QUEUED
        # still waiting because there is no free volume
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id, session=db_session)
        assert request_4['state'] == RequestState.WAITING

        # with maximal volume check -> release one request -> two requests should be released because of attachments
        _delete_requests(mock_scope, [name1, name2, name3, name4], session=db_session)
        request = models.Request(dest_rse_id=dest_rse_id, bytes=2, activity=self.all_activities, state=RequestState.SUBMITTED)
        request.save(session=db_session)
        name1, name2, name3, name4 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2000)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020), 'attributes': {'bytes': 2}},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        dataset_1_name = generate_uuid()
        add_did(mock_scope, dataset_1_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name1, 'scope': mock_scope}], root_account, session=db_session)
        attach_dids(mock_scope, dataset_1_name, [{'name': name2, 'scope': mock_scope}], root_account, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, volume=5, max_transfers=1, session=db_session)
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_grouped_fifo(dest_rse_id, count=1, deadline=0, volume=5, session=db_session)
        # released because it got requested first
        request_1 = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request_1['state'] == RequestState.QUEUED
        # released because the DID is attached to the same dataset
        request_2 = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request_2['state'] == RequestState.QUEUED
        # still waiting because there is no free volume after releasing the two requests above
        request_3 = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request_3['state'] == RequestState.WAITING
        request_4 = get_request_by_did(mock_scope, name4, dest_rse_id, session=db_session)
        assert request_4['state'] == RequestState.WAITING

        # with deadline check -> release 0 requests -> 1 request should be released nonetheless
        _delete_requests(mock_scope, [name1, name2, name3, name4], session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now() - timedelta(hours=2)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now()},
            ]
        )

        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_grouped_fifo(source_rse_id=source_rse_id, count=0, deadline=1, volume=0, session=db_session)
        # queued because of deadline
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        # waiting because count=0
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING

    def test_release_waiting_requests_fifo(self, mock_scope, jdoe_account, db_session, root_account, connected_rse_pair):
        """ REQUEST (CORE): release waiting requests based on FIFO. """
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # without account and activity check
        # two requests -> release one request -> request with oldest requested_at date should be released
        set_rse_transfer_limits(dest_rse_id, self.user_activity, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, 'ignore', max_transfers=1, session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_fifo(dest_rse_id, count=1, session=db_session)
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request2 = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request2['state'] == RequestState.WAITING

        # with activity and account check
        # two requests -> release two request -> requests with correct account and activity should be released
        _delete_requests(mock_scope, [name1, name2], session=db_session)
        name1, name2, name3, name4 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2018),
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'attributes': {'activity': 'ignore'},
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),
                    'account': jdoe_account,
                }, {
                    'source_rse_id': source_rse_id,
                    'dest_rse_id': dest_rse_id,
                    'requested_at': datetime.now().replace(year=2020),  # requested latest but account and activity are correct
                },
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_fifo(dest_rse_id, count=2, account=root_account, activity=self.user_activity, session=db_session)
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING
        request = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING
        request = get_request_by_did(mock_scope, name4, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED

    def test_release_waiting_requests_all(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ REQUEST (CORE): release all waiting requests. """
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        set_rse_transfer_limits(dest_rse_id, self.user_activity, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, max_transfers=1, session=db_session)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2018)},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now().replace(year=2020)},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_all_waiting_requests(dest_rse_id, session=db_session)
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED

    @skiplimitedsql
    def test_release_waiting_requests_per_deadline(self, mock_scope, db_session, root_account, connected_rse_pair):
        """ REQUEST (CORE): release grouped waiting requests that exceeded waiting time."""
        source_rse, source_rse_id, dest_rse, dest_rse_id = connected_rse_pair

        # a request that exceeded the maximal waiting time to be released (1 hour) -> release one request -> only the exceeded request should be released
        set_rse_transfer_limits(dest_rse_id, self.user_activity, max_transfers=1, session=db_session)
        set_rse_transfer_limits(dest_rse_id, self.all_activities, max_transfers=1, session=db_session)
        two_hours = timedelta(hours=2)
        name1, name2 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now() - two_hours},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now()},
            ]
        )
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_per_deadline(source_rse_id=source_rse_id, deadline=1, session=db_session)
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING

        # a request that exceeded the maximal waiting time to be released (1 hour) -> release one request -> release all requests of the same dataset
        name1, name2, name3 = _add_test_replicas_and_request(
            scope=mock_scope, account=root_account, session=db_session,
            request_configs=[
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now() - two_hours},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now()},
                {'source_rse_id': source_rse_id, 'dest_rse_id': dest_rse_id, 'requested_at': datetime.now()},
            ]
        )
        dataset_name = generate_uuid()
        add_did(mock_scope, dataset_name, DIDType.DATASET, root_account, session=db_session)
        attach_dids(mock_scope, dataset_name, [{'name': name1, 'scope': mock_scope}, {'name': name2, 'scope': mock_scope}], root_account, session=db_session)
        preparer.run_once(session=db_session, logger=print)
        db_session.commit()
        release_waiting_requests_per_deadline(source_rse_id=source_rse_id, deadline=1, session=db_session)
        request = get_request_by_did(mock_scope, name1, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name2, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.QUEUED
        request = get_request_by_did(mock_scope, name3, dest_rse_id, session=db_session)
        assert request['state'] == RequestState.WAITING



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import threading
from time import time
from typing import TYPE_CHECKING

import rucio.db.sqla.util
from rucio.common import exception
from rucio.common.config import config_get_list
from rucio.common.exception import RucioException
from rucio.common.logging import setup_logging
from rucio.core import transfer as transfer_core
from rucio.core.request import set_requests_state_if_possible, list_transfer_requests_and_source_replicas
from rucio.core.transfer import prepare_transfers, list_transfer_admin_accounts, build_transfer_paths
from rucio.core.topology import Topology
from rucio.db.sqla.constants import RequestState, RequestType
from rucio.daemons.common import run_daemon

if TYPE_CHECKING:
    from typing import Optional, List
    from sqlalchemy.orm import Session
    from rucio.daemons.common import HeartbeatHandler

graceful_stop = threading.Event()


def stop():
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(once=False, threads=1, sleep_time=10, bulk=100):
    """
    Running the preparer daemon either once or by default in a loop until stop is called.
    """
    setup_logging()

    if rucio.db.sqla.util.is_old_db():
        raise exception.DatabaseException('Database was not updated, daemon won\'t start')

    def preparer_kwargs():
        # not sure if this is needed for threading.Thread, but it always returns a fresh dictionary
        return {'once': once, 'sleep_time': sleep_time, 'bulk': bulk}

    threads = [threading.Thread(target=preparer, name=f'conveyor-preparer-{i}', kwargs=preparer_kwargs(), daemon=True) for i in range(threads)]
    for thr in threads:
        thr.start()

    all_running = True
    while all_running:
        for thr in threads:
            thr.join(timeout=3.14)
            if not thr.is_alive() or graceful_stop.is_set():
                all_running = False
                break

    if graceful_stop.is_set() or once:
        logging.info('conveyor-preparer: gracefully stopping')
    else:
        logging.warning('conveyor-preparer: stopping out of the ordinary')
        graceful_stop.set()

    for thr in threads:
        thr.join(timeout=3.14)

    logging.info('conveyor-preparer: stopped')


def preparer(once, sleep_time, bulk, partition_wait_time=10):
    # Make an initial heartbeat so that all instanced daemons have the correct worker number on the next try
    logger_prefix = executable = 'conveyor-preparer'
    transfertools = config_get_list('conveyor', 'transfertool', False, None)

    run_daemon(
        once=once,
        graceful_stop=graceful_stop,
        executable=executable,
        logger_prefix=logger_prefix,
        partition_wait_time=partition_wait_time,
        sleep_time=sleep_time,
        run_once_fnc=functools.partial(
            run_once,
            transfertools=transfertools,
            bulk=bulk,
        ),
        activities=None,
    )


def run_once(
        transfertools: "Optional[List[str]]" = None,
        bulk: int = 100,
        heartbeat_handler: "Optional[HeartbeatHandler]" = None,
        session: "Optional[Session]" = None,
        **kwargs
) -> bool:
    if heartbeat_handler:
        worker_number, total_workers, logger = heartbeat_handler.live()
    else:
        # This is used in tests
        worker_number, total_workers, logger = 0, 0, logging.log
    if not transfertools:
        transfertools = list(transfer_core.TRANSFERTOOL_CLASSES_BY_NAME)

    start_time = time()
    requests_handled = 0
    try:
        admin_accounts = list_transfer_admin_accounts()
        topology = Topology.create_from_config(logger=logger)
        requests_with_sources = list_transfer_requests_and_source_replicas(
            total_workers=total_workers,
            worker_number=worker_number,
            limit=bulk,
            request_state=RequestState.PREPARING,
            request_type=[RequestType.TRANSFER, RequestType.STAGEIN],
            session=session
        )
        ret = build_transfer_paths(
            topology=topology,
            requests_with_sources=list(requests_with_sources.values()),
            admin_accounts=admin_accounts,
            preparer_mode=True,
            logger=logger,
            session=session,
        )
        requests_handled = sum(len(i) for i in ret)
        if not requests_handled:
            updated_msg = 'had nothing to do'
        else:
            candidate_paths, reqs_no_source, reqs_scheme_mismatch, reqs_only_tape_source, _ = ret
            updated_reqs, reqs_no_transfertool = prepare_transfers(candidate_paths, transfertools=transfertools, logger=logger, session=session)
            updated_msg = f'updated {len(updated_reqs)}/{bulk} requests'

            if reqs_no_transfertool:
                logger(logging.INFO, "Ignoring request because of unsupported transfertool: %s", reqs_no_transfertool)
            reqs_no_source.update(reqs_no_transfertool)
            if reqs_no_source:
                logger(logging.INFO, "Marking requests as no-sources: %s", reqs_no_source)
                set_requests_state_if_possible(reqs_no_source, RequestState.NO_SOURCES, logger=logger)
            if reqs_only_tape_source:
                logger(logging.INFO, "Marking requests as only-tape-sources: %s", reqs_only_tape_source)
                set_requests_state_if_possible(reqs_only_tape_source, RequestState.ONLY_TAPE_SOURCES, logger=logger)
            if reqs_scheme_mismatch:
                logger(logging.INFO, "Marking requests as scheme-mismatch: %s", reqs_scheme_mismatch)
                set_requests_state_if_possible(reqs_scheme_mismatch, RequestState.MISMATCH_SCHEME, logger=logger)
    except RucioException:
        logger(logging.ERROR, 'errored with a RucioException, retrying later', exc_info=True)
        updated_msg = 'errored'
    logger(logging.INFO, '%s, taking %.3f seconds' % (updated_msg, time() - start_time))

    must_sleep = False
    if requests_handled < bulk / 2:
        logger(logging.INFO, "Only %s transfers, which is less than half of the bulk %s", requests_handled, bulk)
        must_sleep = True
    return must_sleep
